@@ -5,7 +5,7 @@
  * object level logging, and fully-qualified object names for all DML and DDL
  * statements where possible (See README.md for details).
  *
- * Copyright (c) 2014-2017, PostgreSQL Global Development Group
+ * Copyright (c) 2014-2020, PostgreSQL Global Development Group
  *------------------------------------------------------------------------------
  */
 #include "postgres.h"
@@ -25,6 +25,7 @@
 #include "miscadmin.h"
 #include "libpq/auth.h"
 #include "nodes/nodes.h"
+#include "nodes/params.h"
 #include "tcop/utility.h"
 #include "tcop/deparse_utility.h"
 #include "utils/acl.h"
@@ -971,7 +972,7 @@ log_select_dml(Oid auditOid, List *rangeTabls)
     foreach(lr, rangeTabls)
     {
         Oid relOid;
-        Relation rel;
+        Oid relNamespaceOid;
         RangeTblEntry *rte = lfirst(lr);
 
         /* We only care about tables, and can ignore subqueries etc. */
@@ -994,13 +995,10 @@ log_select_dml(Oid auditOid, List *rangeTabls)
          * false) then filter out any system relations here.
          */
         relOid = rte->relid;
-        rel = relation_open(relOid, NoLock);
+        relNamespaceOid = get_rel_namespace(relOid);
 
-        if (!auditLogCatalog && IsSystemNamespace(RelationGetNamespace(rel)))
-        {
-            relation_close(rel, NoLock);
+        if (!auditLogCatalog && IsSystemNamespace(relNamespaceOid))
             continue;
-        }
 
         /*
          * Default is that this was not through a grant, to support session
@@ -1059,6 +1057,7 @@ log_select_dml(Oid auditOid, List *rangeTabls)
         switch (rte->relkind)
         {
             case RELKIND_RELATION:
+            case RELKIND_PARTITIONED_TABLE:
                 auditEventStack->auditEvent.objectType = OBJECT_TYPE_TABLE;
                 break;
 
@@ -1097,10 +1096,8 @@ log_select_dml(Oid auditOid, List *rangeTabls)
 
         /* Get a copy of the relation name and assign it to object name */
         auditEventStack->auditEvent.objectName =
-            quote_qualified_identifier(get_namespace_name(
-                                           RelationGetNamespace(rel)),
-                                       RelationGetRelationName(rel));
-        relation_close(rel, NoLock);
+            quote_qualified_identifier(
+                get_namespace_name(relNamespaceOid), get_rel_name(relOid));
 
         /* Perform object auditing only if the audit role is valid */
         if (auditOid != InvalidOid)
@@ -1295,7 +1292,7 @@ pgaudit_ExecutorStart_hook(QueryDesc *queryDesc, int eflags)
 
         /* Initialize the audit event */
         stackItem->auditEvent.commandText = queryDesc->sourceText;
-        stackItem->auditEvent.paramList = queryDesc->params;
+        stackItem->auditEvent.paramList = copyParamList(queryDesc->params);
     }
 
     /* Call the previous hook or standard function */
@@ -1364,11 +1361,30 @@ pgaudit_ProcessUtility_hook(PlannedStmt *pstmt,
         /* Process top level utility statement */
         if (context == PROCESS_UTILITY_TOPLEVEL)
         {
+            /*
+             * If the stack is not empty then the only allowed entries are open
+             * select, show, and explain cursors
+             */
             if (auditEventStack != NULL)
-                elog(ERROR, "pgaudit stack is not empty");
+            {
+                AuditEventStackItem *nextItem = auditEventStack;
+
+                do
+                {
+                    if (nextItem->auditEvent.commandTag != T_SelectStmt &&
+                        nextItem->auditEvent.commandTag != T_VariableShowStmt &&
+                        nextItem->auditEvent.commandTag != T_ExplainStmt)
+                    {
+                        elog(ERROR, "pgaudit stack is not empty");
+                    }
+
+                    nextItem = nextItem->next;
+                }
+                while (nextItem != NULL);
+            }
 
             stackItem = stack_push();
-            stackItem->auditEvent.paramList = params;
+            stackItem->auditEvent.paramList = copyParamList(params);
         }
         else
             stackItem = stack_push();
@@ -1387,6 +1403,19 @@ pgaudit_ProcessUtility_hook(PlannedStmt *pstmt,
             stackItem->auditEvent.commandTag == T_DoStmt &&
             !IsAbortedTransactionBlockState())
             log_audit_event(stackItem);
+
+        /*
+         * A close will free the open cursor which will also free the close
+         * audit entry. Immediately log the close and set stackItem to NULL so
+         * it won't be logged later.
+         */
+        if (stackItem->auditEvent.commandTag == T_ClosePortalStmt)
+        {
+            if (auditLogBitmap & LOG_MISC && !IsAbortedTransactionBlockState())
+                log_audit_event(stackItem);
+
+            stackItem = NULL;
+        }
     }
 
     /* Call the standard process utility chain. */
@@ -1819,7 +1848,7 @@ _PG_init(void)
     if (inited)
         return;
 
-    /* Must be loaded with shared_preload_libaries */
+    /* Must be loaded with shared_preload_libraries */
     if (!process_shared_preload_libraries_in_progress)
         ereport(ERROR, (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
                 errmsg("pgaudit must be loaded via shared_preload_libraries")));
